#!/usr/bin/perl

$debug = 0;
$blockfactor = 1000;

@timenames = qw(REQUEST ENQUEUED DEQUEUED IO COPY XX COMPLETE FINISH);
@opnames   = qw(BFNONE BFOPEN BFCREAT BFREAD BFWRITE BFCLOSE
		BFUNLINK BFSETATTR);

sub usage
{
    my ($errmsg) = @_;
    my ($me) = $0 =~ /([^\/]+)$/;
    $debug && warn caller;
    warn "$errmsg\n" if $errmsg;
    die ("Usage: $me [options] [file] | graph\n",
	 "-r|--raw               raw output, one line per sample\n",
	 "-a|--away              output away stats\n",
	 "-b|--blockfactor=N     each point is an average of N samples\n",
	 "-o|--opname=", join ("|", @opnames[1..(@opnames-1)]), "\n",
	 "-t|--timename=", join ("|", grep {!/REQUEST|XX/} @timenames), "\n",
	 "`graph' is one of GNU plotutils\n");
}

sub main
{
    while ($_ = shift) {
	if (/^-a|--away$/) {
	    $away_mode = 1;
	}
	elsif (/^(-b|--blockfactor=?)(\S*)/) {
	    $blockfactor = $2 || shift;
	}
	elsif (/^(--debug=?)(\S*)/) {
	    $debug = $2 || shift;
	}
	elsif (/^(-o|--op=?)(\S*)/) {
	    my $opname = uc ($2 || shift);
	    usage() unless grep {$_ == $opname} @opnames;
	    usage() if $opname eq BFNONE;
	    $desired_op = array_index ($opname, @opnames);
	    usage("Bad op '$opname'") unless defined $desired_op;
	}
	elsif (/^(-t|--time=?)(\S*)/) {
	    my $timename = uc ($2 || shift);
	    usage() unless grep {$_ == $timename} @timenames;
	    usage() if $timename eq REQUEST;
	    usage() if $timename eq XX;
	    $desired_time = array_index ($timename, @timenames);
	    usage("Bad time '$timename'") unless defined $desired_time;
	}
	elsif (/^-r|--raw$/) {
	    $raw_mode = 1;
	}
	else {
	    unshift @_, $_;
	    last;
	}
    }

    usage() unless $desired_time || $raw_mode || $away_mode;
    usage() if $raw_mode && $away_mode;

    my $file = shift;
    if ($file) {
	open (STDIN, $file) || die "$file: $!\n";
    }

    usage() unless @_ == 0;	# all args processed

    while (read (STDIN, $x, 4+4+4*8)) {
	my ($optype, $away, @times) = unpack "LLL8", $x;

	if ($desired_op && $optype != $desired_op) {
	    next;
	}

	$sample_count++;
	$optype = $opnames[$optype];

	$debug>1 && warn "$optype $away ", join(", ", @times), "\n";

	# fix unfilled times
	for (my $i = 1; $i < @times; $i++) {
	    if ($times[$i] == 0) {
		$times[$i] = $times[$i-1];
	    }
	}

	# output in raw, away, or times mode
	if ($raw_mode) {
	    # convert timemarks to latencies
	    for (my $i = @times; --$i > 0; ) {
		$times[$i] -= $times[$i-1];
	    }
	    print "$optype $away ", join(", ", @times), "\n";
	}
	else {

	    # fight time wrap problem (times[0] is unsigned long
	    # microsecond counter, subject to overflow.  We don't want
	    # the graph to wrap because of this)

	    my $absciss = $wrapstart + $times[0];
	    if ($absciss < $prev_absciss) {
		$wrapstart = $prev_absciss;
		$absciss = $wrapstart + $times[0];
	    }
	    $prev_absciss = $absciss;

	    if ($away_mode) {
		$cum += $away;
	    }
	    else {
		$cum += $times[$desired_time] - $times[$desired_time-1];
# print "++ $cum $times[$desired_time] $times[$desired_time-1] $desired_time";
	    }

	    if ($sample_count % $blockfactor == 0) {
		printf "%.0f %.1f\n", $absciss, $cum/$blockfactor+1;
		$cum = 0;
	    }
	}
    }

    print "\n";			# dataset separator
    print STDERR "$sample_count samples total\n";
    return 0;
}

sub array_index
{
    my ($val, @array) = @_;
    my $index = 0;

    foreach (@array) {
	if ($_ eq $val) {
	    return $index;
	}
	$index++;
    }

    return undef;
}

exit main(@ARGV);
